"""
FootsiesGym - A reinforcement learning environment for HiFight's Footsies game.
"""

from .footsies_env import FootsiesEnv
from . import encoder, typing

__all__ = ["FootsiesEnv", "encoder", "typing"]



from typing import Any
import time 
import numpy as np
from gymnasium import spaces
from ray.rllib import env

from . import encoder, typing
import os
import platform
import socket
import subprocess
import zipfile

from .game import constants, footsies_game
from ..binary_manager import get_binary_manager


class FootsiesEnv(env.MultiAgentEnv):
    metadata = {"render.modes": ["human"]}
    LINUX_ZIP_PATH_HEADLESS = "binaries/footsies_linux_server_021725.zip"
    LINUX_ZIP_PATH_WINDOWED = "binaries/footsies_linux_windowed_021725.zip"
    LINUX_BINARIES_PATH = "binaries/footsies_binaries/footsies.x86_64"
    SPECIAL_CHARGE_FRAMES = 60
    GUARD_BREAK_REWARD = 0.3

    observation_space = spaces.Dict(
        {
            agent: spaces.Box(
                low=-np.inf,
                high=np.inf,
                shape=(encoder.FootsiesEncoder.observation_size,),
            )
            for agent in ["p1", "p2"]
        }
    )

    action_space = spaces.Dict(
        {
            agent: spaces.Discrete(
                len(
                    [
                        constants.EnvActions.NONE,
                        constants.EnvActions.BACK,
                        constants.EnvActions.FORWARD,
                        constants.EnvActions.ATTACK,
                        constants.EnvActions.BACK_ATTACK,
                        constants.EnvActions.FORWARD_ATTACK,
                        # NOTE(chase): This is a special input that holds down
                        # attack for 60 frames. It's just too long of a sequence
                        # to easily learn by holding ATTACK for so long.
                        constants.EnvActions.SPECIAL_CHARGE,
                    ]
                )
            )
            for agent in ["p1", "p2"]
        }
    )

    def __init__(self, config: dict[Any, Any] = None):
        super(FootsiesEnv, self).__init__()

        if config is None:
            config = {}
        self.config = config
        self.use_build_encoding = config.get("use_build_encoding", False)
        self.agents: list[typing.AgentID] = ["p1", "p2"]
        self.possible_agents: list[typing.AgentID] = self.agents.copy()
        self._agent_ids: set[typing.AgentID] = set(self.agents)

        self.evaluation = config.get("evaluation", False)

        self.t: int = 0
        self.max_t: int = config.get("max_t", 1000)
        self.frame_skip = config.get("frame_skip", 4)
        observation_delay = config.get("observation_delay", 16)

        assert (
            observation_delay % self.frame_skip == 0
        ), "observation_delay must be divisible by frame_skip"

        self.encoder = encoder.FootsiesEncoder(
            observation_delay=observation_delay // self.frame_skip
        )


        
        port = config.get("port", None)
        self.headless = config.get("headless", True)
        # we'll start game servers at 50051 for training
        # and 40051 for evaluation. Worker index starts at
        # 1, so we won't see 40050/50050.
        if port is None:
            start_port = 40050 if self.evaluation else 50050
            port = (
                start_port
                + int(config.worker_index)
                * config.get("num_envs_per_worker", 1)
                + config.get("vector_index", 0)
            )

        # If specified, we'll launch the binaries from the environment itself.
        self.server_process = None
        launch_binaries = config.get("launch_binaries", False)
        if launch_binaries:
            self._launch_binaries(port=port)

        self.game = footsies_game.FootsiesGame(
            host=config.get("host", "localhost"),
            port=port,
        )

        self.last_game_state = None
        self.special_charge_queue = {
            "p1": -1,
            "p2": -1,
        }

    def _is_port_in_use(self, port: int) -> bool:
        """Check if a port is already in use."""
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
            try:
                sock.bind(('localhost', port))
                return False
            except OSError:
                return True

    def _launch_binaries(self, port: int):
        # Check if we're on a supported platform
        if platform.system().lower() in ["windows", "darwin"]:
            raise RuntimeError(
                "Binary launching is only supported on Linux. "
                "Please launch the footsies server manually or use a Linux system."
            )

        # Check to ensure the linux binaries exist in binaries/footsies_binaries/footsies.x86_64
        
        project_root = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
        binary_path = os.path.join(project_root, "binaries", "footsies_binaries", "footsies.x86_64")

        if not os.path.exists(binary_path):
            # Use binary manager to download and extract binaries from Git LFS
            binary_manager = get_binary_manager()
            
            # Ensure binaries are available (download if needed)
            if not binary_manager.ensure_binaries_available("linux"):
                raise FileNotFoundError(
                    "Failed to download footsies binaries from Git LFS. "
                    "Please check your internet connection and try again."
                )
            
            # Get the appropriate zip file path
            zip_path = binary_manager.get_binary_path("linux", windowed=not self.headless)
            if not zip_path or not zip_path.exists():
                raise FileNotFoundError(
                    f"Could not find {'headless' if self.headless else 'windowed'} Linux binary after download."
                )
            
            print(f"Extracting footsies binaries from {zip_path}...")
            
            # Extract the zip file to the package binaries directory
            binaries_dir = os.path.join(project_root, "binaries")
            os.makedirs(binaries_dir, exist_ok=True)
            
            with zipfile.ZipFile(zip_path, 'r') as zip_ref:
                zip_ref.extractall(binaries_dir)
            
            # Find the extracted folder and rename it to footsies_binaries
            extracted_items = os.listdir(binaries_dir)
            extracted_folder = None
            
            # Look for a folder that was just extracted (not the existing footsies_binaries)
            for item in extracted_items:
                item_path = os.path.join(binaries_dir, item)
                if (os.path.isdir(item_path) and 
                    item != "footsies_binaries" and 
                    not item.endswith(".zip") and
                    "footsies" in item.lower()):
                    extracted_folder = item
                    break
            
            if extracted_folder:
                old_path = os.path.join(binaries_dir, extracted_folder)
                new_path = os.path.join(binaries_dir, "footsies_binaries")
                
                # Remove existing footsies_binaries if it exists
                if os.path.exists(new_path):
                    import shutil
                    shutil.rmtree(new_path)
                
                os.rename(old_path, new_path)
                print(f"Renamed {extracted_folder} to footsies_binaries")
            
            # Verify the binary now exists
            if not os.path.exists(binary_path):
                raise FileNotFoundError(
                    f"Failed to extract or find footsies binary at {binary_path} after extraction."
                )
        
        # We'll also want to make sure the binary is executable
        if not os.access(binary_path, os.X_OK):
            # If not, make it executable
            os.chmod(binary_path, 0o755)

        # Check if the port is already in use
        if self._is_port_in_use(port):
            print(f"Port {port} is already in use. Skipping binary launch.")
            return

        command = [binary_path, "--port", str(port)]
        
        # Launch the process in the background. We can store the process if we need to manage it later.
        self.server_process = subprocess.Popen(command, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
        print(f"Launched footsies binary on port {port}.")
        time.sleep(5)

    def close(self):
        """Clean up resources when the environment is closed."""
        if hasattr(self, 'server_process') and self.server_process is not None:
            try:
                self.server_process.terminate()
                # Give it a moment to terminate gracefully
                self.server_process.wait(timeout=5)
                print(f"Terminated footsies server process (PID: {self.server_process.pid}).")
            except subprocess.TimeoutExpired:
                # Force kill if it doesn't terminate gracefully
                self.server_process.kill()
                self.server_process.wait()
                print(f"Force killed footsies server process (PID: {self.server_process.pid}).")
            except Exception as e:
                print(f"Error terminating server process: {e}")
            finally:
                self.server_process = None

    def __del__(self):
        """Ensure cleanup happens when the object is garbage collected."""
        self.close()


    def get_obs(self, game_state):
        if self.use_build_encoding:
            encoded_state = self.game.get_encoded_state()
            encoded_state_dict = {
                "p1": np.asarray(
                    encoded_state.player1_encoding, dtype=np.float32
                ),
                "p2": np.asarray(
                    encoded_state.player2_encoding, dtype=np.float32
                ),
            }
            return encoded_state_dict
        else:
            return self.encoder.encode(game_state)

    def reset(
        self,
        *,
        seed: int | None = None,
        options: dict | None = None,
    ) -> tuple[
        dict[typing.AgentID, typing.ObsType], dict[typing.AgentID, Any]
    ]:
        """Resets the environment to the starting state
        and returns the initial observations for all agents.

        :return: Tuple of observations and infos for each agent.
        :rtype: tuple[dict[typing.AgentID, typing.ObsType], dict[typing.AgentID, Any]]
        """
        self.t = 0
        self.game.reset_game()
        self.game.start_game()

        self.encoder.reset()

        if not self.use_build_encoding:
            self.last_game_state = self.game.get_state()

        observations = self.get_obs(self.last_game_state)

        return observations, {agent: {} for agent in self.agents}

    def step(self, actions: dict[typing.AgentID, typing.ActionType]) -> tuple[
        dict[typing.AgentID, typing.ObsType],
        dict[typing.AgentID, float],
        dict[typing.AgentID, bool],
        dict[typing.AgentID, bool],
        dict[typing.AgentID, dict[str, Any]],
    ]:
        """Step the environment with the provided actions for all agents.

        :param actions: Dictionary mapping agent ids to their actions for this step.
        :type actions: dict[typing.AgentID, typing.ActionType]
        :return: Tuple of observations, rewards, terminates, truncateds and infos for all agents.
        :rtype: tuple[ dict[typing.AgentID, typing.ObsType], dict[typing.AgentID, float], dict[typing.AgentID, bool], dict[typing.AgentID, bool], dict[typing.AgentID, dict[str, Any]], ]
        """
        self.t += 1

        for agent_id in self.agents:
            empty_queue = self.special_charge_queue[agent_id] < 0
            action_is_special_charge = (
                actions[agent_id] == constants.EnvActions.SPECIAL_CHARGE
            )

            # Refill the charge queue only if we're not already in a special charge.
            if action_is_special_charge and empty_queue:
                self.special_charge_queue[agent_id] = (
                    self._build_charged_special_queue()
                )

            if self.special_charge_queue[agent_id] >= 0:
                self.special_charge_queue[agent_id] -= 1
                actions[agent_id] = self._convert_to_charge_action(
                    actions[agent_id]
                )

        p1_action = self.game.action_to_bits(actions["p1"], is_player_1=True)
        p2_action = self.game.action_to_bits(actions["p2"], is_player_1=False)

        game_state = self.game.step_n_frames(
            p1_action=p1_action, p2_action=p2_action, n_frames=self.frame_skip
        )
        observations = self.get_obs(game_state)

        terminated = game_state.player1.is_dead or game_state.player2.is_dead

        # Zero-sum game: 1 if other player is dead, -1 if you're dead:
        rewards = {
            "p1": int(game_state.player2.is_dead)
            - int(game_state.player1.is_dead),
            "p2": int(game_state.player1.is_dead)
            - int(game_state.player2.is_dead),
        }

        if self.config.get("reward_guard_break", False):
            p1_prev_guard_health = self.last_game_state.player1.guard_health
            p2_prev_guard_health = self.last_game_state.player2.guard_health
            p1_guard_health = game_state.player1.guard_health
            p2_guard_health = game_state.player2.guard_health

            if p2_guard_health < p2_prev_guard_health:
                rewards["p1"] += self.GUARD_BREAK_REWARD
                rewards["p2"] -= self.GUARD_BREAK_REWARD
            if p1_guard_health < p1_prev_guard_health:
                rewards["p2"] += self.GUARD_BREAK_REWARD
                rewards["p1"] -= self.GUARD_BREAK_REWARD

        terminateds = {
            "p1": terminated,
            "p2": terminated,
            "__all__": terminated,
        }

        truncated = self.t >= self.max_t
        truncateds = {
            "p1": truncated,
            "p2": truncated,
            "__all__": truncated,
        }

        self.last_game_state = game_state

        # encoded_state = self.game.get_encoded_state()
        # encoded_state_dict = {
        #     "p1": np.asarray(
        #         encoded_state.player1_encoding, dtype=np.float32
        #     ),
        #     "p2": np.asarray(
        #         encoded_state.player2_encoding, dtype=np.float32
        #     ),
        # }

        # for a_id, ob in observations.items():
        #     matched_obs = np.isclose(ob, encoded_state_dict[a_id]).all()
        #     assert matched_obs

        return observations, rewards, terminateds, truncateds, self.get_infos()

    def get_infos(self):
        return {agent: {} for agent in self.agents}

    def _build_charged_special_queue(self):
        assert self.SPECIAL_CHARGE_FRAMES % self.frame_skip == 0
        steps_to_apply_attack = int(
            self.SPECIAL_CHARGE_FRAMES // self.frame_skip
        )
        return steps_to_apply_attack

    @staticmethod
    def _convert_to_charge_action(action: int) -> int:
        if action == constants.EnvActions.BACK:
            return constants.EnvActions.BACK_ATTACK
        elif action == constants.EnvActions.FORWARD:
            return constants.EnvActions.FORWARD_ATTACK
        else:
            return constants.EnvActions.ATTACK

    def _build_charged_queue_features(self):
        return {
            "p1": {
                "special_charge_queue": self.special_charge_queue["p1"]
                / self.SPECIAL_CHARGE_FRAMES
            },
            "p2": {
                "special_charge_queue": self.special_charge_queue["p2"]
                / self.SPECIAL_CHARGE_FRAMES
            },
        }



"""
FootsiesGym - A reinforcement learning environment for HiFight's Footsies game.

This package provides a Gymnasium-compatible environment for training reinforcement
learning agents on the Footsies fighting game.

Binaries are automatically downloaded from Git LFS when first needed.
"""

from .footsies.footsies_env import FootsiesEnv
from .footsies import encoder, typing
from .binary_manager import get_binary_manager

__version__ = "0.1.1"
__all__ = ["FootsiesEnv", "encoder", "typing", "make"]

# Initialize binary manager (but don't download yet - wait until needed)
_binary_manager = get_binary_manager()


def make(config: dict | None = None, platform: str = "linux", launch_binaries: bool = True):
    """
    Create a FootsiesGym environment.
    
    Args:
        config: Configuration dictionary for the environment
        platform: Platform to run on (currently only "linux" supported for auto-launch)
        launch_binaries: Whether to automatically launch game binaries
        
    Returns:
        FootsiesEnv: The configured environment instance
    """
    if launch_binaries:
        assert platform == "linux", (
            "Only linux is supported for automated binary launching. "
            "Create the environment manually and launch binaries by hand to use MacOS. "
            "Windows TBD."
        )
    
    default_config = {
        "platform": platform, 
        "launch_binaries": launch_binaries,
        "max_t": 4000,
        "frame_skip": 4,
        "observation_delay": 16,
        "reward_guard_break": False,
    }

    if config is not None:
        default_config.update(config)

    return FootsiesEnv(config=default_config)



"""
Binary manager for FootsiesGym - handles automatic binary downloads.
"""

import os
import urllib.request
import urllib.error
from pathlib import Path
from typing import Optional


class BinaryManager:
    """Manages automatic downloading and caching of FootsiesGym binaries."""
    
    # Direct download URLs for binary files
    # Using GitHub raw files as the primary source
    DOWNLOAD_BASE_URL = "https://github.com/cmcdonald/FootsiesGym/raw/main/footsiesgym/binaries"
    
    # Fallback URLs in case primary fails
    FALLBACK_URLS = [
        "https://raw.githubusercontent.com/cmcdonald/FootsiesGym/main/footsiesgym/binaries",
    ]
    
    BINARY_FILES = {
        "footsies_linux_server_021725.zip": "footsies_linux_server_021725.zip",
        "footsies_linux_windowed_021725.zip": "footsies_linux_windowed_021725.zip", 
        "footsies_mac_headless_5709b6d.zip": "footsies_mac_headless_5709b6d.zip",
        "footsies_mac_windowed_5709b6d.zip": "footsies_mac_windowed_5709b6d.zip",
    }
    
    def __init__(self):
        self.package_dir = Path(__file__).parent
        self.binaries_dir = self.package_dir / "binaries"
        self.binaries_dir.mkdir(exist_ok=True)
    
    def ensure_binaries_available(self, platform: str = "linux") -> bool:
        """
        Ensure the required binaries are available for the given platform.
        Downloads them automatically if they don't exist locally.
        
        Args:
            platform: Target platform ("linux" or "mac")
            
        Returns:
            bool: True if binaries are available, False otherwise
        """
        required_files = self._get_required_files(platform)
        
        # Check if all required files exist
        missing_files = []
        for filename in required_files:
            file_path = self.binaries_dir / filename
            if not file_path.exists():
                missing_files.append(filename)
        
        if not missing_files:
            print(f"✅ All {platform} binaries are available")
            return True
        
        print(f"📥 Downloading missing {platform} binaries: {missing_files}")
        
        # Download missing files automatically
        success = True
        for filename in missing_files:
            if not self._download_binary(filename):
                success = False
        
        return success
    
    def _get_required_files(self, platform: str) -> list[str]:
        """Get the list of required binary files for a platform."""
        if platform.lower() == "linux":
            return [
                "footsies_linux_server_021725.zip",
                "footsies_linux_windowed_021725.zip"
            ]
        elif platform.lower() == "mac":
            return [
                "footsies_mac_headless_5709b6d.zip", 
                "footsies_mac_windowed_5709b6d.zip"
            ]
        else:
            return []
    
    def _download_binary(self, filename: str) -> bool:
        """
        Get a binary file by copying from local repository or downloading from HTTP.
        
        Args:
            filename: Name of the file to download
            
        Returns:
            bool: True if successful, False otherwise
        """
        local_path = self.binaries_dir / filename
        
        # First, try to find the file in the local repository
        # Look for it in the parent directory structure
        repo_binaries_path = self.package_dir.parent / "footsiesgym" / "binaries" / filename
        
        if repo_binaries_path.exists():
            try:
                print(f"  Copying {filename} from local repository...")
                import shutil
                shutil.copy2(repo_binaries_path, local_path)
                print(f"  ✅ Copied {filename} ({local_path.stat().st_size} bytes)")
                return True
            except Exception as e:
                print(f"  ❌ Failed to copy from local repository: {e}")
        
        # If local copy fails, try HTTP download as fallback
        urls_to_try = [self.DOWNLOAD_BASE_URL] + self.FALLBACK_URLS
        
        for base_url in urls_to_try:
            url = f"{base_url}/{filename}"
            
            try:
                print(f"  Downloading {filename} from {base_url}...")
                
                # Create request with user agent to avoid blocking
                request = urllib.request.Request(url)
                request.add_header('User-Agent', 'FootsiesGym/0.1.0')
                
                with urllib.request.urlopen(request, timeout=30) as response:
                    # Check if we got a valid response
                    if response.status == 200:
                        with open(local_path, 'wb') as f:
                            # Download in chunks to handle large files
                            while True:
                                chunk = response.read(8192)
                                if not chunk:
                                    break
                                f.write(chunk)
                        
                        print(f"  ✅ Downloaded {filename} ({local_path.stat().st_size} bytes)")
                        return True
                    else:
                        print(f"  ❌ HTTP {response.status} from {base_url}")
                        continue
                        
            except urllib.error.HTTPError as e:
                print(f"  ❌ HTTP error {e.code} from {base_url}: {e.reason}")
                continue
            except urllib.error.URLError as e:
                print(f"  ❌ URL error from {base_url}: {e.reason}")
                continue
            except Exception as e:
                print(f"  ❌ Unexpected error from {base_url}: {e}")
                continue
        
        print(f"  ❌ Failed to get {filename} from all sources")
        return False
    

    
    def get_binary_path(self, platform: str = "linux", windowed: bool = False) -> Optional[Path]:
        """
        Get the path to the appropriate binary file.
        
        Args:
            platform: Target platform ("linux" or "mac")
            windowed: Whether to get windowed version (True) or headless (False)
            
        Returns:
            Path to the binary file, or None if not available
        """
        if platform.lower() == "linux":
            filename = "footsies_linux_windowed_021725.zip" if windowed else "footsies_linux_server_021725.zip"
        elif platform.lower() == "mac":
            filename = "footsies_mac_windowed_5709b6d.zip" if windowed else "footsies_mac_headless_5709b6d.zip"
        else:
            return None
        
        binary_path = self.binaries_dir / filename
        return binary_path if binary_path.exists() else None


# Global instance
_binary_manager = None

def get_binary_manager() -> BinaryManager:
    """Get the global binary manager instance."""
    global _binary_manager
    if _binary_manager is None:
        _binary_manager = BinaryManager()
    return _binary_manager



import logging
import time

import grpc
import numpy as np

from experimentation.utils import utils
from footsiesgym.footsies.game import constants
from footsiesgym.footsies.game.proto import footsies_service_pb2 as footsies_pb2
from footsiesgym.footsies.game.proto import footsies_service_pb2_grpc as footsies_pb2_grpc

log = logging.getLogger(__name__)

TIMEOUT_DURATION_S = 5


class FootsiesGame:
    """The FootsiesGame class establishes communication between the
    game server and the Python harness via gRPC. It provides methods
    to start the game, reset it, get the current state, and step the
    game by a certain number of frames.
    """

    def __init__(self, host: str = "localhost", port: int = "50051"):
        self.host = host
        self.port = port
        self.stub = self._initialize_stub()

    @utils.timeout(TIMEOUT_DURATION_S)
    def _initialize_stub(self) -> footsies_pb2_grpc.FootsiesGameServiceStub:
        try:
            channel = grpc.insecure_channel(f"{self.host}:{self.port}")
            return footsies_pb2_grpc.FootsiesGameServiceStub(channel)
        except Exception as e:
            log.error(f"Error connecting to gRPC stub with exception: {e}")
            raise e

    @utils.timeout(TIMEOUT_DURATION_S)
    def start_game(self) -> None:
        """Starts the game by calling the StartGame RPC."""
        try:
            self.stub.StartGame(footsies_pb2.Empty())

            while not self.is_ready():
                log.info("Game not ready...")
                time.sleep(0.5)
            log.info("StartGame called successfully")

        except Exception as e:
            log.error(f"Error calling StartGame with exception: {e}")
            raise e

    @utils.timeout(TIMEOUT_DURATION_S)
    def is_ready(self) -> bool:
        """Checks if the game is ready by calling the IsReady RPC."""
        try:
            return self.stub.IsReady(footsies_pb2.Empty()).value
        except Exception as e:
            log.error(f"Error calling IsReady with exception: {e}")
            raise e

    @utils.timeout(TIMEOUT_DURATION_S)
    def get_state(self) -> footsies_pb2.GameState:
        """Gets the current game state by calling the GetState RPC."""
        try:
            return self.stub.GetState(footsies_pb2.Empty())
        except Exception as e:
            log.error(f"Error calling GetState with exception: {e}")
            raise e

    @utils.timeout(TIMEOUT_DURATION_S)
    def get_encoded_state(self) -> footsies_pb2.EncodedGameState:
        """Gets the current encoded game state by calling the GetEncodedState RPC."""
        try:
            return self.stub.GetEncodedState(footsies_pb2.Empty())
        except Exception as e:
            log.error(f"Error calling GetEncodedState with exception: {e}")
            raise e

    @utils.timeout(TIMEOUT_DURATION_S)
    def step_n_frames(
        self, p1_action: int, p2_action: int, n_frames: int
    ) -> footsies_pb2.GameState:
        """Steps the game by n_frames with the given player actions. The provided actions will be repeated for all n_frames."""
        try:
            step_input = footsies_pb2.StepInput(
                p1_action=p1_action, p2_action=p2_action, nFrames=n_frames
            )
            return self.stub.StepNFrames(step_input)
        except Exception as e:
            log.error(f"Error calling StepNFrames with exception: {e}")
            raise e

    @utils.timeout(TIMEOUT_DURATION_S)
    def reset_game(self) -> None:
        """Resets the game by calling the ResetGame RPC."""
        try:
            self.stub.ResetGame(footsies_pb2.Empty())
        except Exception as e:
            log.error(f"Error calling ResetGame with exception: {e}")
            raise e

    @staticmethod
    def action_to_bits(action: int, is_player_1: bool) -> int:
        """Converts an action to its corresponding bit representation."""

        if isinstance(action, np.ndarray):
            action = action.item()

        if is_player_1:
            if action == constants.EnvActions.BACK:
                action = constants.GameActions.LEFT
            elif action == constants.EnvActions.FORWARD:
                action = constants.GameActions.RIGHT
            elif action == constants.EnvActions.BACK_ATTACK:
                action = constants.GameActions.LEFT_ATTACK
            elif action == constants.EnvActions.FORWARD_ATTACK:
                action = constants.GameActions.RIGHT_ATTACK
        else:
            if action == constants.EnvActions.BACK:
                action = constants.GameActions.RIGHT
            elif action == constants.EnvActions.FORWARD:
                action = constants.GameActions.LEFT
            elif action == constants.EnvActions.BACK_ATTACK:
                action = constants.GameActions.RIGHT_ATTACK
            elif action == constants.EnvActions.FORWARD_ATTACK:
                action = constants.GameActions.LEFT_ATTACK

        return constants.ACTION_TO_BITS[action]



#!/usr/bin/env python3
"""
Setup script for FootsiesGym package.
This is a fallback setup.py for compatibility with older pip versions.
The main configuration is in pyproject.toml.
"""

from setuptools import setup

if __name__ == "__main__":
    setup()



import dataclasses


@dataclasses.dataclass
class EnvActions:
    NONE = 0
    BACK = 1
    FORWARD = 2
    ATTACK = 3
    BACK_ATTACK = 4
    FORWARD_ATTACK = 5
    SPECIAL_CHARGE = 6


@dataclasses.dataclass
class GameActions:
    NONE = 0
    LEFT = 1
    RIGHT = 2
    ATTACK = 3
    LEFT_ATTACK = 4
    RIGHT_ATTACK = 5


@dataclasses.dataclass
class ActionBits:
    NONE: int = 0
    LEFT: int = 1 << 0
    RIGHT: int = 1 << 1
    ATTACK: int = 1 << 2
    LEFT_ATTACK: int = LEFT | ATTACK
    RIGHT_ATTACK: int = RIGHT | ATTACK


ACTION_TO_BITS = {
    GameActions.NONE: ActionBits.NONE,
    GameActions.LEFT: ActionBits.LEFT,
    GameActions.RIGHT: ActionBits.RIGHT,
    GameActions.ATTACK: ActionBits.ATTACK,
    GameActions.LEFT_ATTACK: ActionBits.LEFT_ATTACK,
    GameActions.RIGHT_ATTACK: ActionBits.RIGHT_ATTACK,
}


@dataclasses.dataclass
class ActionID:
    STAND = 0
    FORWARD = 1
    BACKWARD = 2
    DASH_FORWARD = 10
    DASH_BACKWARD = 11
    N_ATTACK = 100
    B_ATTACK = 105
    N_SPECIAL = 110
    B_SPECIAL = 115
    DAMAGE = 200
    GUARD_M = 301
    GUARD_STAND = 305
    GUARD_CROUCH = 306
    GUARD_BREAK = 310
    GUARD_PROXIMITY = 350
    DEAD = 500
    WIN = 510


FOOTSIES_ACTION_IDS = {
    "STAND": ActionID.STAND,
    "FORWARD": ActionID.FORWARD,
    "BACKWARD": ActionID.BACKWARD,
    "DASH_FORWARD": ActionID.DASH_FORWARD,
    "DASH_BACKWARD": ActionID.DASH_BACKWARD,
    "N_ATTACK": ActionID.N_ATTACK,
    "B_ATTACK": ActionID.B_ATTACK,
    "N_SPECIAL": ActionID.N_SPECIAL,
    "B_SPECIAL": ActionID.B_SPECIAL,
    "DAMAGE": ActionID.DAMAGE,
    "GUARD_M": ActionID.GUARD_M,
    "GUARD_STAND": ActionID.GUARD_STAND,
    "GUARD_CROUCH": ActionID.GUARD_CROUCH,
    "GUARD_BREAK": ActionID.GUARD_BREAK,
    "GUARD_PROXIMITY": ActionID.GUARD_PROXIMITY,
    "DEAD": ActionID.DEAD,
    "WIN": ActionID.WIN,
}



import collections
import copy
from typing import Any

import numpy as np

from footsiesgym.footsies.game import constants
from footsiesgym.footsies.game.proto import footsies_service_pb2 as footsies_pb2


class EncoderMethods:
    @staticmethod
    def one_hot(
        value: int | float | str, collection: list[int | float | str]
    ) -> np.ndarray:
        vector = np.zeros(len(collection), dtype=np.float32)
        vector[collection.index(value)] = 1
        return vector


class FootsiesEncoder:
    """Encoder class to generate observations from the game state"""

    observation_size: int = 81

    def __init__(self, observation_delay: int):
        self._encoding_history = {
            agent_id: collections.deque(maxlen=int(observation_delay))
            for agent_id in ["p1", "p2"]
        }
        self.observation_delay = observation_delay
        self._last_common_state: np.ndarray | None = None

    def reset(self):
        self._encoding_history = {
            agent_id: collections.deque(maxlen=int(self.observation_delay))
            for agent_id in ["p1", "p2"]
        }

    def encode(
        self,
        game_state: footsies_pb2.GameState,
        **kwargs,
    ) -> dict[str, Any]:
        """Encodes the game state into observations for all agents.

        kwargs can be used to pass in additional features that
        are added directly to the observation, keyed by the agent
        IDs, e.g.,
            kwargs = {
                "p1": {"p1_feature": 1},
                "p2": {"p2_feature": 2},
            }

        :param game_state: The game state to encode
        :type game_state: footsies_pb2.GameState
        :return: The encoded observations for all agents.
        :rtype: dict[str, Any]
        """
        common_state = self.encode_common_state(game_state)
        p1_encoding = self.encode_player_state(
            game_state.player1, game_state.frame_count, **kwargs.get("p1", {})
        )
        p2_encoding = self.encode_player_state(
            game_state.player2, game_state.frame_count, **kwargs.get("p2", {})
        )

        observation_delay = min(
            self.observation_delay, len(self._encoding_history["p1"])
        )

        if observation_delay > 0:
            p1_delayed_encoding = self._encoding_history["p1"][
                -observation_delay
            ]
            p2_delayed_encoding = self._encoding_history["p2"][
                -observation_delay
            ]
        else:
            p1_delayed_encoding = copy.deepcopy(p1_encoding)
            p2_delayed_encoding = copy.deepcopy(p2_encoding)

        self._encoding_history["p1"].append(p1_encoding)
        self._encoding_history["p2"].append(p2_encoding)
        self._last_common_state = common_state

        # Create features dictionary and export to JSON
        features = {}
        current_index = 0

        # Common state
        features["common_state"] = {
            "start": current_index,
            "length": len(common_state),
        }
        current_index += len(common_state)

        # Concatenate the observations for the undelayed encoding
        p1_encoding = np.hstack(list(p1_encoding.values()), dtype=np.float32)
        p2_encoding = np.hstack(list(p2_encoding.values()), dtype=np.float32)

        # Concatenate the observations for the delayed encoding
        p1_delayed_encoding = np.hstack(
            list(p1_delayed_encoding.values()), dtype=np.float32
        )
        p2_delayed_encoding = np.hstack(
            list(p2_delayed_encoding.values()), dtype=np.float32
        )

        p1_centric_observation = np.hstack(
            [common_state, p1_encoding, p2_delayed_encoding]
        )

        p2_centric_observation = np.hstack(
            [common_state, p2_encoding, p1_delayed_encoding]
        )

        return {"p1": p1_centric_observation, "p2": p2_centric_observation}

    def get_last_encoding(self) -> dict[str, np.ndarray]:
        if self._last_common_state is None:
            return None

        return {
            "common_state": self._last_common_state.reshape(-1),
            "p1": np.hstack(
                list(self._encoding_history["p1"][-1].values()),
                dtype=np.float32,
            ),
            "p2": np.hstack(
                list(self._encoding_history["p2"][-1].values()),
                dtype=np.float32,
            ),
        }

    def encode_common_state(
        self, game_state: footsies_pb2.GameState
    ) -> np.ndarray:
        p1_state, p2_state = game_state.player1, game_state.player2

        dist_x = (
            np.abs(p1_state.player_position_x - p2_state.player_position_x)
            / 8.0
        )

        return np.array(
            [
                dist_x,
            ],
            dtype=np.float32,
        )

    def encode_player_state(
        self,
        player_state: footsies_pb2.PlayerState,
        frame_count: int,
        **kwargs,
    ) -> dict[str, int | float | list]:
        """Encodes the player state into observations.

        :param player_state: The player state to encode
        :type player_state: footsies_pb2.PlayerState
        :return: The encoded observations for the player
        :rtype: dict[str, Any]

        TODO(chase): Test mirroring the positions so
            the agent always thinks it's LHS
        """
        feature_dict = {
            "player_position_x": player_state.player_position_x / 4.0,
            "velocity_x": player_state.velocity_x / 5.0,
            "is_dead": int(player_state.is_dead),
            "vital_health": player_state.vital_health,
            "guard_health": EncoderMethods.one_hot(
                player_state.guard_health, [0, 1, 2, 3]
            ),
            "current_action_id": self._encode_action_id(
                player_state.current_action_id
            ),
            "current_action_frame": player_state.current_action_frame / 25,
            "current_action_frame_count": player_state.current_action_frame_count
            / 25,
            "current_action_remaining_frames": (
                player_state.current_action_frame_count
                - player_state.current_action_frame
            )
            / 25,
            "is_action_end": int(player_state.is_action_end),
            "is_always_cancelable": int(player_state.is_always_cancelable),
            "current_action_hit_count": player_state.current_action_hit_count,
            "current_hit_stun_frame": player_state.current_hit_stun_frame / 10,
            "is_in_hit_stun": int(player_state.is_in_hit_stun),
            "sprite_shake_position": player_state.sprite_shake_position,
            "max_sprite_shake_frame": player_state.max_sprite_shake_frame / 10,
            "is_face_right": int(player_state.is_face_right),
            "current_frame_advantage": player_state.current_frame_advantage
            / 10,
            # The below features leak some information about the opponent!
            "would_next_forward_input_dash": int(
                player_state.would_next_forward_input_dash
            ),
            "would_next_backward_input_dash": int(
                player_state.would_next_backward_input_dash
            ),
            "special_attack_progress": min(
                player_state.special_attack_progress, 1.0
            ),
        }

        if kwargs:
            feature_dict.update(kwargs)

        return feature_dict

    def _encode_action_id(self, action_id: int) -> np.ndarray:
        """Encodes the action id into a one-hot vector.

        :param action_id: The action id to encode
        :type action_id: int
        :return: The encoded one-hot vector
        :rtype: np.ndarray
        """

        action_id_values = list(constants.FOOTSIES_ACTION_IDS.values())
        action_vector = np.zeros(len(action_id_values), dtype=np.float32)

        # Get the index of the action id in constants.ActionID
        action_index = action_id_values.index(action_id)
        action_vector[action_index] = 1

        assert action_vector.max() == 1 and action_vector.min() == 0

        return action_vector

    def _encode_input_buffer(
        self, input_buffer: list[int], last_n: int | None = None
    ) -> np.ndarray:
        """Encodes the input buffer into a one-hot vector.

        :param input_buffer: The input buffer to encode
        :type input_buffer: list[int]
        :return: The encoded one-hot vector
        :rtype: np.ndarray
        """

        if last_n is not None:
            input_buffer = input_buffer[last_n:]

        ib_encoding = []
        for action_id in input_buffer:
            arr = [0] * (len(constants.ACTION_TO_BITS) + 1)
            arr[action_id] = 1
            ib_encoding.extend(arr)

        input_buffer_vector = np.asarray(ib_encoding, dtype=np.float32)

        return input_buffer_vector
